/**
    Copyright (C) 2001, 2002, 2003, 2004  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#ifndef _TC2_H
#define _TC2_H

#include <pthread.h>
#include <tctypes.h>
#include <tclist.h>
#include <tcconf.h>
#include <stdarg.h>

typedef struct {
    char *interface;
    char *xname; /* Exported name */
    char *iname; /* Internal name */
} tc2_export;

typedef struct {
    char *name;
    void *data;
    size_t size;
} tc2_extdata;

typedef struct _tc2_interface tc2_interface;
typedef struct _tc2_module tc2_module;

typedef struct {
    char *name;           /* Name              */
    unsigned int version; /* Version           */
    char *description;    /* Description       */
    char **interfaces;    /* Implemented interfaces */
    tc2_export *exports;  /* Symbol exportlist */
    tc2_extdata *extdata;
    tc2_module *this_module;
} tc2_module_info;

struct _tc2_interface {
    char *name;
    tc2_module *loaded;
    tclist_t *modules;
};

struct _tc2_module {
    char *filename;              /* Filename          */
    void *handle;                /* Handle            */
    tc2_module_info *module;     /* Module info       */
    tc2_interface **interfaces;
    char *option;
    tclist_t *dep_list;
    int autoclean;
    int loaded;
    pthread_cond_t cond;
    pthread_mutex_t mut;
    pthread_t init_thread;
};

/* Add directory to module path. */
extern int tc2_add_modpath(char *dir);

/* Add configuration file. */
extern int tc2_add_config(char *filename);

/* Get a configuration section. */
extern tcconf_section_t *tc2_get_conf(char *sec);

/* Initialize most things, returns 0 on success. */
extern int tc2_init(void);

/* Run the main loop. */
extern int tc2_run(void);

/* Free resources allocated by TC2 */
extern int tc2_free(void);

extern int tc2_print(char *tag, int level, const char *, ...);

#define TC2_PRINT_ERROR   0
#define TC2_PRINT_WARNING 1
#define TC2_PRINT_INFO    2
#define TC2_PRINT_VERBOSE 3
#define TC2_PRINT_DEBUG   (TC2_PRINT_VERBOSE + 4)

typedef int (*tc2_print_t)(const char *tag, int level, const char *, va_list);
extern int tc2_setprint(char *tag, int, int, char *);
extern int tc2_regprint(char *name, tc2_print_t);
extern int tc2_delprint(char *name);

#define TC2_PRINT_NONE ((tc2_print_t *)1L)

extern void tc2_verbose(int v);
extern void tc2_debug(int d);

/* Get a function pointer */
extern void *_tc2_get_symbol(tc2_module *module,
			     char *interface, char *sym);
#define tc2_get_symbol(interface, sym)			\
    _tc2_get_symbol(THIS_MODULE, interface, sym)

/* Send a request to the kernel. */
extern int _tc2_request(int type, tc2_module *module, int wait, ...);
#define tc2_request(type, wait, args...)			\
    _tc2_request(type, THIS_MODULE, wait , ## args)

extern int tc2_call_hook(char *hook, void *opt);
extern int tc2_add_hook(char *hook, int(*fn)(void *));
extern int tc2_rm_hook(char *hook, int(*fn)(void *));

#ifndef THIS_MODULE
#define THIS_MODULE NULL
#endif

/* Kernel request types. */
#define TC2_LOAD_MODULE        0x0001
#define TC2_UNLOAD_MODULE      0x0002
#define TC2_UNLOAD_ALL         0x0003
#define TC2_ABORT              0x0004
#define TC2_PANIC              0x0005
#define TC2_AUTOCLEAN          0x0006
#define TC2_GET_MODULES        0x0007
#define TC2_ADD_DEPENDENCY     0x0008
#define TC2_DEL_DEPENDENCY     0x0009
#define TC2_DEL_ALL_DEPS       0x000a
#define TC2_GET_MODULE         0x000b
#define TC2_ADD_MODULE         0x000c

#endif /* _TC2_H */
