/**
    Copyright (C) 2001, 2002, 2003, 2004  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include <stdio.h>
#include <dlfcn.h>
#include <tcdirent.h>
#include <stdlib.h>
#include <unistd.h>
#include <tcstring.h>
#include <semaphore.h>
#include <tclist.h>
#include <tcconf.h>
#include <tcalloc.h>
#include <tc2.h>
#include <kernel.h>

struct mod_dir {
    char *dirname;
    struct dirent **dir;
    int files;
};

/* Selector for scandir */
static int
ismodule(const struct dirent *f)
{
    if(strstr(f->d_name, ".so") == 0)
	return 0;
    if(strchr(f->d_name, '-') == 0)
	return 0;

    return 1;
}

/* This global is only used during initial module scanning */
static int n_interfaces;

static int
get_module_priority(char *interface, char *module)
{
    tcconf_section_t *sec;
    int priority = 0;
    char *m, *i;
    void *cs = NULL;

    kprint("init", TC2_PRINT_DEBUG + 1,
	   "Checking priority for interface:%s module:%s\n",
 	   interface, module);

    if(!(sec = tc2_get_conf("TC2/interface")))
	return 0;

    do {
	int tp;
	if(tcconf_nextvalue(sec, "priority", &cs, "%s %s %d",
			    &i, &m, &tp) == 3){
	    if(!strcmp(interface, i) && !strcmp(module, m)){
		if(tp > priority){
		    priority = tp;
		}
	    }
	    free(i);
	    free(m);
	}
    } while(cs);

    tcfree(sec);
    return priority;
}

extern int
tc2_add_module(char *file, tc2_module_info *md)
{
    int j, k;
    tc2_module *mod = calloc(1, sizeof(tc2_module));
    tc2_module *tmp_mod;

    mod->filename = file;

    mod->module = calloc(1, sizeof(tc2_module_info));
    mod->module->name = strdup(md->name);
    mod->module->version = md->version;
    mod->module->description = strdup(md->description);

    /* Count elements in exportlist */
    for(j = 0; md->exports[j].xname != NULL; j++);

    /* Allocate memory for export tclist_t */
    kprint("init", TC2_PRINT_DEBUG + 4,
	   "Allocating memory for export list. (%d exports)\n", j);
    mod->module->exports = malloc(sizeof(tc2_export)*(j+1));

    /* Copy export list from module */
    for(j = 0; md->exports[j].xname != NULL; j++) {
	mod->module->exports[j].interface = strdup(md->exports[j].interface);
	mod->module->exports[j].xname = strdup(md->exports[j].xname);
	mod->module->exports[j].iname = strdup(md->exports[j].iname);
    }

    /* Terminate export tclist_t */
    mod->module->exports[j].xname = NULL;
    mod->module->exports[j].iname = NULL;

    for(j = 0; md->interfaces[j]; j++);
    mod->interfaces = calloc(j + 1, sizeof(tc2_interface *));
    mod->module->interfaces = calloc(j + 1, sizeof(char *));
    for(j = 0; md->interfaces[j]; j++){
	char *i = md->interfaces[j];
	int replace;

	for(k = 0; interfaces[k] && strcmp(interfaces[k]->name, i); k++);

	if(k == n_interfaces - 1){
	    int ni = n_interfaces * 2;
	    interfaces = realloc(interfaces, ni * sizeof(*interfaces));
	    memset(interfaces + n_interfaces, 0,
		   n_interfaces * sizeof(*interfaces));
	    n_interfaces = ni;
	}

	if(!interfaces[k]){
	    interfaces[k] = calloc(1, sizeof(tc2_interface));
	    interfaces[k]->name = strdup(i);
	    interfaces[k]->modules = tclist_new(TC_LOCK_SLOPPY);
	}

	tclist_push(interfaces[k]->modules, mod);

	mod->interfaces[j] = interfaces[k];
	mod->module->interfaces[j] = interfaces[k]->name;

	if(!(replace = tchash_search(modules, i, -1, mod, &tmp_mod))){
	    if(!strcmp(mod->module->name, tmp_mod->module->name)){
		if(mod->module->version >= tmp_mod->module->version)
		    replace = 1;
		kprint("init", TC2_PRINT_DEBUG + 3,
		       "%s: replace %s version %d with version %d: %s\n", 
		       i, mod->module->name, tmp_mod->module->version,
		       mod->module->version, replace? "yes": "no");
	    } else {
		int modp = get_module_priority(i, mod->module->name);
		int tmp_modp = get_module_priority(i, tmp_mod->module->name);

		if(modp >= tmp_modp)
		    replace = 1;

		kprint("init", TC2_PRINT_DEBUG + 3,
		       "%s: replace %s (pri %d) with %s (pri %d)? %s\n", 
		       i, tmp_mod->module->name, tmp_modp, mod->module->name,
		       modp, replace? "yes": "no");
	    }
	}

	if(replace){
	    int l;
	    tchash_replace(modules, i, -1, mod, NULL);
	    for(l = 0; md->exports[l].xname != NULL; l++) {
		if(!strcmp(md->exports[l].interface, i)){
		    char *s = MKSYM(md->exports[l].interface,
				    md->exports[l].xname);
		    tchash_replace(avail_syms, s, -1, mod, NULL);
		    free(s);
		}
	    }
	}
    }

    mod->dep_list = tclist_new(TC_LOCK_SLOPPY);
    pthread_cond_init(&mod->cond, NULL);
    pthread_mutex_init(&mod->mut, NULL);

    if(!tchash_search(module_names, mod->module->name, -1, mod, &tmp_mod)){
	if(mod->module->version >= tmp_mod->module->version){
	    tchash_replace(module_names, mod->module->name, -1, mod, NULL);
	    tc2_free_module(tmp_mod);
	}
    }

    return 0;
}

static int
read_config(void)
{
    tcconf_section_t *sec;
    void *cs = NULL;
    char *dir;
    char *tag, *log;
    int start, end;
    int n;

    sec = tc2_get_conf("TC2");

    while(n = tcconf_nextvalue(sec, "module_path", &cs, "%s", &dir), cs){
	if(n == 1){
	    tc2_add_modpath(dir);
	}
    }

    while(n = tcconf_nextvalue(sec, "print", &cs, "%s%i%i%s",
			       &tag, &start, &end, &log), cs){
	if(n == 4){
	    tc2_setprint(tag, start, end, log);
	}
    }

    tcconf_getvalue(sec, "unload_modules", "%i", &unload_modules);

    tcfree(sec);
    return 0;
}

extern int
tc2_init(void)
{
    void *handle;
    tc2_module_info *mod_info;
    char *error=NULL;
    tclist_t *dirlist;
    tclist_item_t *de = NULL;
    int k, files = 0;
    struct mod_dir *mdir;
    char *dir;

    tc2_init_print();
    read_config();

    /* Initialize the event queue. */
    kreq_init();

    dirlist = tclist_new(TC_LOCK_SLOPPY);

    /* Get dirlisting */
    for(dir = tclist_next(module_dirs, &de); de != NULL;
	dir = tclist_next(module_dirs, &de)){
	int n;
	struct mod_dir *md = alloca(sizeof(struct mod_dir));
	struct dirent **modlist;

	kprint("init", TC2_PRINT_VERBOSE + 1,
	       "Searching for modules in '%s'\n", dir);
	n = scandir(dir, &modlist, ismodule, NULL);
	if (n < 0) {
	    perror(dir);
	    continue;
	}
	md->dirname = dir;
	md->dir = modlist;
	md->files = n;
	tclist_push(dirlist, md);
	files += n;
    }

    /* Create hash tables */
    avail_syms = tchash_new(files * 15, TC_LOCK_STRICT, 0);
    loaded_syms = tchash_new(files * 15, TC_LOCK_STRICT, 0);
    modules = tchash_new(files * 2, TC_LOCK_STRICT, 0);
    module_names = tchash_new(files * 2, TC_LOCK_STRICT, 0);

    /* Allocate memory for module data */
    n_interfaces = 2*files;
    kprint("init", TC2_PRINT_DEBUG + 4,
	   "Allocating memory for a maximum of %d module pointers.\n",
	   n_interfaces);
    interfaces = calloc(n_interfaces, sizeof(tc2_interface *));

    while((mdir = tclist_shift(dirlist)) != NULL){
	struct dirent **modlist = mdir->dir;
	for(k = 0; k < mdir->files; k++) {
	    char *buf;

	    buf = malloc(strlen(mdir->dirname) +
			 strlen(modlist[k]->d_name) + 2);
	    sprintf(buf, "%s/%s",
		    mdir->dirname, modlist[k]->d_name);

	    /* Open a module */
	    kprint("init", TC2_PRINT_DEBUG,
		   "Loading info for module %s.\n", buf);
	    handle = dlopen(buf, RTLD_LAZY);
	    if ((error = dlerror()) != NULL) {
		kprint("init", TC2_PRINT_ERROR,
		       "error opening module '%s': %s\n", buf, error);
		free(buf);
		continue;
	    }

	    /* Get the info struct */
	    mod_info = dlsym(handle,"tc2__module_info");
	    if ((error = dlerror()) != NULL)  {
		kprint("init", TC2_PRINT_ERROR,
		       "Module '%s' is missing tc2__module_info: %s\n",
		       buf, error);
		dlclose(handle);
		free(buf);
		continue;
	    }

	    tc2_add_module(buf, mod_info);
	    
	    dlclose(handle);
	    free(modlist[k]);
	}
	free(modlist);
    }

    tclist_destroy(dirlist, NULL);

    tc2_init_hook();

    return 0;
}

